
/* Chrysalide - Outil d'analyse de fichiers binaires
 * array.c - manipulations de tableaux génériques
 *
 * Copyright (C) 2010-2013 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "array.h"


#include "../expression-int.h"



/* Définition d'un accès à une cellule de tableau (instance) */
struct _GArrayAccess
{
    GDecExpression parent;                  /* A laisser en premier        */

    GDecExpression *array;                  /* Elément auquel accéder      */
    GDecExpression *index;                  /* Position de la cellule      */

};


/* Définition d'un accès à une cellule de tableau (classe) */
struct _GArrayAccessClass
{
    GDecExpressionClass parent;             /* A laisser en premier        */

};



/* Initialise la classe des accès aux cellules de tableau. */
static void g_array_access_class_init(GArrayAccessClass *);

/* Initialise une instance d'accès à une cellule de tableau. */
static void g_array_access_init(GArrayAccess *);

/* Visite un ensemble hiérarchique d'instructions décompilées. */
static bool g_array_access_visit(GArrayAccess *, dec_instr_visitor_cb, DecInstrVisitFlags, void *);

/* Remplace une instruction décompilée par une autre. */
static bool g_array_access_replace(GArrayAccess *, GDecInstruction *, GDecInstruction *);

/* Imprime pour l'écran un version humaine d'une expression. */
static GBufferLine *g_array_access_print(const GArrayAccess *, GCodeBuffer *, GBufferLine *, GLangOutput *);



/* Indique le type défini pour un accès à une cellule de tableau. */
G_DEFINE_TYPE(GArrayAccess, g_array_access, G_TYPE_DEC_EXPRESSION);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des accès aux cellules de tableau.      *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_array_access_class_init(GArrayAccessClass *klass)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr = instance à initialiser.                               *
*                                                                             *
*  Description : Initialise une instance d'accès à une cellule de tableau.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_array_access_init(GArrayAccess *expr)
{
    GDecInstruction *instr;                 /* Autre version de l'objet    */

    instr = G_DEC_INSTRUCTION(expr);

    instr->visit = (dec_instr_visit_fc)g_array_access_visit;
    instr->replace = (dec_instr_replace_fc)g_array_access_replace;
    instr->print = (dec_instr_print_fc)g_array_access_print;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : array = tableau auquel on doit accéder.                      *
*                index = indice de la cellule concernée.                      *
*                                                                             *
*  Description : Construit un accès à une cellule de tableau comme expression.*
*                                                                             *
*  Retour      : Expression mise en place.                                    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GDecInstruction *g_array_access_new(GDecExpression *array, GDecExpression *index)
{
    GArrayAccess *result;                 /* Expression à retourner      */

    result = g_object_new(G_TYPE_ARRAY_ACCESS, NULL);

    result->array = array;
    result->index = index;

    return G_DEC_INSTRUCTION(result);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr     = première instruction à venir visiter.             *
*                callback = procédure à appeler à chaque instruction visitée. *
*                flags    = moments des appels à réaliser en retour.          *
*                data     = données quelconques associées au visiteur.        *
*                                                                             *
*  Description : Visite un ensemble hiérarchique d'instructions décompilées.  *
*                                                                             *
*  Retour      : true si le parcours a été jusqu'à son terme, false sinon.    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_array_access_visit(GArrayAccess *expr, dec_instr_visitor_cb callback, DecInstrVisitFlags flags, void *data)
{
    bool result;                            /* Bilan à retourner           */

    result = _g_dec_instruction_visit(G_DEC_INSTRUCTION(expr->array), G_DEC_INSTRUCTION(expr),
                                      callback, flags, data);

    if (result)
        result = _g_dec_instruction_visit(G_DEC_INSTRUCTION(expr->index), G_DEC_INSTRUCTION(expr),
                                          callback, flags, data);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr = première instruction à venir ausculter.               *
*                old  = instruction décompilée à venir remplacer.             *
*                new  = instruction décompilée à utiliser dorénavant.         *
*                                                                             *
*  Description : Remplace une instruction décompilée par une autre.           *
*                                                                             *
*  Retour      : true si un remplacement a été effectué, false sinon.         *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_array_access_replace(GArrayAccess *expr, GDecInstruction *old, GDecInstruction *new)
{
    bool result;                            /* Bilan à retourner           */

    if (expr->array == G_DEC_EXPRESSION(old))
    {
        g_object_unref(G_OBJECT(expr->array));
        g_object_ref(G_OBJECT(new));
        expr->array = G_DEC_EXPRESSION(new);

        result = true;

    }
    else
        result = g_dec_instruction_replace(G_DEC_INSTRUCTION(expr->array), old, new);

    if (expr->index == G_DEC_EXPRESSION(old))
    {
        g_object_unref(G_OBJECT(expr->index));
        g_object_ref(G_OBJECT(new));
        expr->index = G_DEC_EXPRESSION(new);

        result = true;

    }
    else
        result |= g_dec_instruction_replace(G_DEC_INSTRUCTION(expr->index), old, new);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr   = expression à transcrire en version humaine.         *
*                buffer = tampon où doit se réaliser l'insertion.             *
*                line   = ligne d'impression prête à emploi ou NULL.          *
*                output = langage de programmation de sortie.                 *
*                                                                             *
*  Description : Imprime pour l'écran un version humaine d'une expression.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static GBufferLine *g_array_access_print(const GArrayAccess *expr, GCodeBuffer *buffer, GBufferLine *line, GLangOutput *output)
{
    GBufferLine *result;                    /* Ligne à retourner           */

    result = g_dec_instruction_print(G_DEC_INSTRUCTION(expr->array),
                                     buffer, line, output);

    g_buffer_line_append_text(result, BLC_ASSEMBLY_HEAD, "[", 1, RTT_RAW, NULL);

    result = g_dec_instruction_print(G_DEC_INSTRUCTION(expr->index),
                                     buffer, result, output);

    g_buffer_line_append_text(result, BLC_ASSEMBLY_HEAD, "]", 1, RTT_RAW, NULL);

    return result;

}
